"""
answerQuestion.py This is the last step in the Collins and Quillian experiment. 

Like assocStatic,  read in all 9 sets of synapses for the animal, operation
and property cells.

Use spike sources to turn on the question, and providedAnswer.  The
question turns on animal and operation  (or just the animal for an
isA).

The animal and operation turn on associated properties.  These then
turn on generatedAnswer.  Equal compares genrated and providedAnswer.
If they're equal it turns on the outputCells 0 CA.  The timer will turn
ouputCells 1 on roughly 150ms after the question is asked.

This can do the association (tests commented out) and at least
one hierarchical relation.    This tests 7 of the 8 questions (it's just
one instance of the question, we should have several).  It gets 4 of 7
right (missing pigeon isa animal; canary can fly; and canary has skin).
"""

import os #for printSynapticWeightsToFile
import pyNN.nest as sim
from pyNN.nest import native_synapse_type
from pyNN.nest import native_cell_type
import nealParams as nealParameters
from nealCoverClass import NealCoverFunctions
from stateMachineClass import FSAHelperFunctions
from timerClass import TimerClass 

#constants
numAnimalCells = 200   #(animal (amphibian) 
                      #(fish shark salmon bass pike) #7
                      #(bird canary ostrich robin goose pigeon) 
                      #(mammal dog cat rat bear monkey human))  #20
numOperationCells = 50 #eats, breathes, is, has, can
numPropertyCells = 50  #(properties (can) move fly sing
                       #            (is) yellow (has) skin feathers gills)
    #canary is yellow
    #bird can fly
    #animal has skin
    #fish breathes withGills

numAnswerCells = 250 #for provided, generated and equal
animalAnswerStart = 50  #animal, fish, bird, mammal

import sys
import cPickle as pickle

def printPklSpikes(fileName,outFileName):
        fileHandle = open(fileName)
        outFileHandle = open(outFileName,'w')
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    outFileHandle.write(str( neuronNum) + " " + str(spikes[spike])+ "\n")
        outFileHandle.close()
        fileHandle.close()

def printSynapticWeights(synapses):
    weights = synapses.get(["weight"],format="array")
    print weights

def printSynapticWeightsToFile(synapses,fileName):
    fileHandle = open(fileName,'w')
    weights = synapses.get(["weight"],format="array")
    weightArray = weights[0]
    for preSynapticNeuronNumber in range (0,len(weightArray)):
        synapsesFromNeuron = weightArray[preSynapticNeuronNumber]
        fileHandle.write(str(preSynapticNeuronNumber))
        fileHandle.write(" ")
        fileHandle.write(str(synapsesFromNeuron))
        fileHandle.write("\n")
    fileHandle.write("\n")
    fileHandle.close()

def makeQuestionAnswerSynapses(animalCells, animalStartNeuron,
                               operationCells, operationStartNeuron,
                               providedAnswerCells,providedAnswerStartNeuron,
                               timer, timeStart):
    presentArray = {'spike_times': [timeStart]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.3

    #start the timer
    timer.connectSpikeGenToStart(spikeSource)

    #stimulate provideAnswer
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = providedAnswerStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, providedAnswerCells,connector,'excitatory')

    #stimulate animal
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = animalStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, animalCells,connector,'excitatory')

    #stimulate operation
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = operationStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, operationCells,connector,'excitatory')

def makeHierQuestionAnswerSynapses(animalCells, animalStartNeuron,
                                   providedAnswerCells,
                                   providedAnswerStartNeuron,primeCells,
                                   timer, timeStart):
    presentArray = {'spike_times': [timeStart]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.3

    #start the timer
    timer.connectSpikeGenToStart(spikeSource)

    #stimulate provideAnswer
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = providedAnswerStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, providedAnswerCells,connector,'excitatory')

    #stimulate animal
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = animalStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, animalCells,connector,'excitatory')

    #turn on prime
    fsa.turnOnStateFromSpikeSource(spikeSource,primeCells,0)

def makeTestSynapses(animalCells,operationCells,providedAnswerCells,primeCells,
                     timer):
    makeHierQuestionAnswerSynapses(animalCells, 120,
                               providedAnswerCells,120,primeCells,
                               timer,100) #pigeon is -> bird

    makeQuestionAnswerSynapses(animalCells, 80,operationCells, 20,
                               providedAnswerCells,0,
                               timer,400) #canary is -> withGills (false)

    makeQuestionAnswerSynapses(animalCells, 80,operationCells, 20,
                               providedAnswerCells,30,
                               timer,700) #canary is -> yellow

    makeQuestionAnswerSynapses(animalCells, 80,operationCells, 40,
                               providedAnswerCells,10,
                               timer,1000) #canary can -> fly

    makeHierQuestionAnswerSynapses(animalCells, 180,
                               providedAnswerCells,50,primeCells,
                               timer,1300) #monkey is -> animal

    makeHierQuestionAnswerSynapses(animalCells, 120,
                               providedAnswerCells,70,primeCells,
                               timer,1600) #pigeon is -> fish (false)

    makeQuestionAnswerSynapses(animalCells, 80,operationCells, 30,
                               providedAnswerCells,40,
                               timer,1900) #canary has -> skin (false)

    makeHierQuestionAnswerSynapses(animalCells, 120,
                               providedAnswerCells,170,primeCells,
                               timer,2200) #pigeon is -> pigeon


def makeStops(animalCells,operationCells,propertyCells,providedAnswerCells,
              generatedAnswerCells,answersEqualCells,primeCells,outputCells):
    presentArray = {'spike_times': [300,600,900,1200,1500,1800,2100,2400]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = -2.4 
    connector = []
    for toOffset in range (0,numOperationCells):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, operationCells, connector,'inhibitory')
    neal.nealProjection(spikeSource, propertyCells, connector,'inhibitory')

    for toOffset in range (numOperationCells,numAnimalCells):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, animalCells, connector,'inhibitory')

    numCAs = numAnswerCells/10
    for CA in range (0,numCAs):
        fsa.oneNeuronTurnsOffState(spikeSource,0,providedAnswerCells,CA)
        fsa.oneNeuronTurnsOffState(spikeSource,0,generatedAnswerCells,CA)
        fsa.oneNeuronTurnsOffState(spikeSource,0,answersEqualCells,CA)

    fsa.oneNeuronTurnsOffState(spikeSource,0,outputCells,0)
    fsa.oneNeuronTurnsOffState(spikeSource,0,outputCells,1)
    fsa.oneNeuronTurnsOffState(spikeSource,0,primeCells,0)


def setupAnswerCAs(providedAnswerCells,generatedAnswerCells,equalCells):
    numCAs = numAnswerCells/10
    for CA in range (0,numCAs):
        fsa.makeCA(providedAnswerCells,CA)
        fsa.makeCA(generatedAnswerCells,CA)
        fsa.makeCA(equalCells,CA)

def makePropertyToGeneratedAnswerSynapses(propertyCells,generatedAnswerCells):
    weight = 0.01
    numCAs = numPropertyCells/10

    connector = []
    for CA in range (0,numCAs):
        for fromOffset in range (0,10):
            fromNeuron = (CA*10)+fromOffset
            for toOffset in range (0,10):
                toNeuron = (CA*10) + toOffset
                connector = connector + [(fromNeuron,toNeuron,weight,
                                          nealParameters.DELAY)]
    neal.nealProjection(propertyCells, generatedAnswerCells,connector,
                        'excitatory')

def connectAnimalToGeneratedAnswer(animalCells,fromStart,generatedAnswerCells,
                                   toStart):
    weight = 0.01
    connector = []
    for fromOffset in range (0,5):
        fromNeuron = fromStart+fromOffset
        for toOffset in range (0,10):
            toNeuron = toStart + toOffset
            connector = connector + [(fromNeuron,toNeuron,weight,
                                          nealParameters.DELAY)]
    neal.nealProjection(animalCells, generatedAnswerCells,connector,
                        'excitatory')

def makeAnimalToGeneratedAnswerSynapses(animalCells,generatedAnswerCells):
    for animal in range (0,19):
        animalCellNumber = animal*10
        answerCellNumber = animalCellNumber + animalAnswerStart
        connectAnimalToGeneratedAnswer(animalCells,animalCellNumber,
                                       generatedAnswerCells,answerCellNumber)

#connect each of the input CAs to its equal CA with a half.
def connectAnswers(generatedCells,providedCells,equalCells,outputCells):
    print 'connect'
    numCAs = numAnswerCells/10
    for CA in range (0,numCAs):
        fsa.stateHalfTurnsOnState(providedCells,CA,equalCells,CA)
        fsa.stateHalfTurnsOnState(generatedCells,CA,equalCells,CA)
        fsa.stateTurnsOnState(equalCells,CA,outputCells,0)
        fsa.stateTurnsOffState(outputCells,0,equalCells,CA)

#output is a two state machine.  The first state prevents the second.
def setOutputSynapses(outputCells):
    fsa.makeCA(outputCells,0)
    fsa.makeCA(outputCells,1)
    fsa.stateTurnsOffState(outputCells,0,outputCells,1)

def primeAnimal(animalCells,startNeuron,primeCells):
    primeWeight = 0.0005
    connector = []
    for fromNeuron in range (0,10):
        for toOffset in range (0,5):
                toNeuron = startNeuron + toOffset
                connector = connector + [(fromNeuron,toNeuron,primeWeight,
                                          nealParameters.DELAY)]
    neal.nealProjection(primeCells, animalCells,connector,'excitatory')

#note that the animalCells aren't from the state class
def makePrimeHierarchySynapses(animalCells,primeCells):
    primeAnimal(animalCells,0,primeCells)#animal
    primeAnimal(animalCells,20,primeCells)#fish
    primeAnimal(animalCells,70,primeCells)#bird
    primeAnimal(animalCells,130,primeCells)#mammal


#---Read in the learned synapses.
def skipBlanks(inputArray,numberOfCurrentInput):
    while (inputArray[numberOfCurrentInput] == ''):
            numberOfCurrentInput += 1
    return numberOfCurrentInput

def readSynapseFile(fileName):
    connector = []
    preSynapticNeuronNumber = -1
    postSynapticNeuronNumber = 0
    needNeuronNumber = True
    inputFile = open(fileName,'r')
    inputLine = inputFile.readline()
    #while inputLine != "":
    while (len(inputLine) > 1): #go until a blank line
        inputLine = inputLine.strip() #remove the \n
        #print inputLine
        inputs = inputLine.split(' ')
        numInputs = len(inputs)
        currentInput = 0
        if (needNeuronNumber):
            needNeuronNumber = False
            postSynapticNeuronNumber = 0
            temp = int(inputs[0])
            preSynapticNeuronNumber = preSynapticNeuronNumber + 1
            if (preSynapticNeuronNumber != temp):
                    print 'bad neuron number'
            currentInput = 3

        #see if the line has the last synapse
        if (']' in inputLine):
            needNeuronNumber = True
            #remove the ] from the last input
            lastInput =  inputs[numInputs-1]
            lastInput = lastInput[0:len(lastInput)-1]
            inputs[numInputs-1] = lastInput
        
        #process one line
        while (currentInput < numInputs):
            currentInput = skipBlanks(inputs,currentInput)
            #print inputs[currentInput]
            if (inputs[currentInput] == 'nan'):
                synapticWeight = 0.0
            else:
                synapticWeight = float(inputs[currentInput])
            #print preSynapticNeuronNumber,postSynapticNeuronNumber,synapticWeight
            if (synapticWeight > 0.00001):
            #if ((synapticWeight > 0.00001) and (preSynapticNeuronNumber != postSynapticNeuronNumber)):
                    connector = connector + [(preSynapticNeuronNumber,postSynapticNeuronNumber,synapticWeight,1.0)]
            postSynapticNeuronNumber = postSynapticNeuronNumber + 1
            currentInput = currentInput + 3

        inputLine = inputFile.readline()
    inputFile.close()
    
    return connector

def readSynapses(animalCells,numAnimalCells,operationCells,numOperationCells,
                 propertyCells,numPropertyCells):
    connector = readSynapseFile("results/storedDoubleAnimalWeights.txt")
    print "intra animal synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    intraAnimalSynapses = sim.Projection(animalCells,animalCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedOperationWeights.txt")
    print "intra operation synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    intraOperationSynapses = sim.Projection(operationCells,operationCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedPropertyWeights.txt")
    print "intra property synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    intraPropertySynapses = sim.Projection(propertyCells,propertyCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedAnimalToOperationWeights.txt")
    print "animal to operation synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    animalToOperationSynapses = sim.Projection(animalCells,operationCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedAnimalToPropertyWeights.txt")
    print "animal to property synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    animalToPropertySynapses = sim.Projection(animalCells,propertyCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedOperationToAnimalWeights.txt")
    print "operation to animal synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    operationToAnimalSynapses = sim.Projection(operationCells,animalCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedOperationToPropertyWeights.txt")
    print "operation to property synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    operationToPropertySynapses = sim.Projection(operationCells,propertyCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedPropertyToAnimalWeights.txt")
    print "property to animal synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    propertyToAnimalSynapses = sim.Projection(propertyCells,animalCells, 
                connectorList, sim.StaticSynapse()) 

    connector = readSynapseFile("results/storedPropertyToOperationWeights.txt")
    print "property to operation synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    propertyToOperationSynapses = sim.Projection(propertyCells,operationCells, 
                connectorList, sim.StaticSynapse()) 

    return [intraAnimalSynapses,intraOperationSynapses,intraPropertySynapses,
            animalToOperationSynapses, animalToPropertySynapses,
            operationToAnimalSynapses, operationToPropertySynapses,
            propertyToAnimalSynapses, propertyToOperationSynapses]


def printStaticSynapticWeights(synapseSets):
        
    printSynapticWeightsToFile(synapseSets[0],"results/animalWeights.txt")
    printSynapticWeightsToFile(synapseSets[1],"results/operationWeights.txt")
    printSynapticWeightsToFile(synapseSets[2],"results/propertyWeights.txt")
    printSynapticWeightsToFile(synapseSets[3],
                               "results/animalToOperationWeights.txt")
    printSynapticWeightsToFile(synapseSets[4],
                               "results/animalToPropertyWeights.txt")
    printSynapticWeightsToFile(synapseSets[5],
                               "results/operationToAnimalWeights.txt")
    printSynapticWeightsToFile(synapseSets[6],
                               "results/operationToPropertyWeights.txt")
    printSynapticWeightsToFile(synapseSets[7],
                               "results/propertyToAnimalWeights.txt")
    printSynapticWeightsToFile(synapseSets[8],
                               "results/propertyToOperationWeights.txt")

#--main-
sim.nest.Install("chrismodule")
CompenseIFCondExpNeuronType = native_cell_type('compenseIF_cond_exp')
#print sim.nest.GetDefaults('compenseIF_cond_exp')
neal = NealCoverFunctions(nealParameters.simulator)
fsa = FSAHelperFunctions(nealParameters.simulator)

sim.setup(timestep=1.0, max_delay=1.0)

#Create the Neurons
AnimalCells = sim.Population(numAnimalCells,CompenseIFCondExpNeuronType, 
        {'weight_std':21.0,'tau_syn_ex':5.0,'g_L':20.0,'tau_syn_in' : 2.0})
AnimalCells.record(['spikes'])
OperationCells = sim.Population(numOperationCells,CompenseIFCondExpNeuronType, 
                                {'weight_std':20.0,'tau_syn_ex':5.0,'g_L':20.0})
OperationCells.record(['spikes'])
PropertyCells = sim.Population(numPropertyCells,CompenseIFCondExpNeuronType, 
                                {'weight_std':20.0,'tau_syn_ex':5.0,'g_L':20.0})
PropertyCells.record(['spikes'])

#params from stateMachine
providedAnswerCells = sim.Population(numAnswerCells,sim.IF_cond_exp,
                                     fsa.CELL_PARAMS)
providedAnswerCells.record(['spikes'])

generatedAnswerCells = sim.Population(numAnswerCells,sim.IF_cond_exp,
                                     fsa.CELL_PARAMS)
generatedAnswerCells.record(['spikes'])

answersEqualCells = sim.Population(numAnswerCells,sim.IF_cond_exp,
                                     fsa.CELL_PARAMS)
answersEqualCells.record(['spikes'])

outputCells = sim.Population(20,sim.IF_cond_exp,fsa.CELL_PARAMS)
outputCells.record(['spikes'])

hierPrimeCells = sim.Population(10,sim.IF_cond_exp,fsa.CELL_PARAMS)
hierPrimeCells.record(['spikes'])

outputTimer = TimerClass(nealParameters.simulator,10)

#---Make the synapses
makeTestSynapses(AnimalCells,OperationCells, providedAnswerCells,hierPrimeCells,
                 outputTimer)
makeStops(AnimalCells,OperationCells,PropertyCells,providedAnswerCells,
          generatedAnswerCells,answersEqualCells,hierPrimeCells,outputCells)

synapseSets = readSynapses(AnimalCells,numAnimalCells,OperationCells,
                           numOperationCells,PropertyCells,numPropertyCells)

setupAnswerCAs(providedAnswerCells,generatedAnswerCells,answersEqualCells)
makePropertyToGeneratedAnswerSynapses(PropertyCells,generatedAnswerCells)
makeAnimalToGeneratedAnswerSynapses(AnimalCells,generatedAnswerCells)
connectAnswers(generatedAnswerCells,providedAnswerCells,answersEqualCells,
               outputCells)
setOutputSynapses(outputCells)
fsa.makeCA(hierPrimeCells,0)
makePrimeHierarchySynapses(AnimalCells,hierPrimeCells)

outputTimer.makeStopTimerSynapses()
outputTimer.timerStartsState(outputCells,1)


sim.run(2500)

AnimalCells.write_data('results/animalCells.pkl','spikes')
OperationCells.write_data('results/operationCells.pkl','spikes')
PropertyCells.write_data('results/propertyCells.pkl','spikes')
printPklSpikes('results/animalCells.pkl','results/animalCells.sp')
printPklSpikes('results/operationCells.pkl','results/operationCells.sp')
printPklSpikes('results/propertyCells.pkl','results/propertyCells.sp')

providedAnswerCells.write_data('results/providedAnswerCells.pkl','spikes')
printPklSpikes('results/providedAnswerCells.pkl',
               'results/providedAnswerCells.sp')
generatedAnswerCells.write_data('results/generatedAnswerCells.pkl','spikes')
printPklSpikes('results/generatedAnswerCells.pkl',
               'results/generatedAnswerCells.sp')
answersEqualCells.write_data('results/answersEqualCells.pkl','spikes')
printPklSpikes('results/answersEqualCells.pkl',
               'results/answersEqualCells.sp')
outputCells.write_data('results/outputCells.pkl','spikes')
printPklSpikes('results/outputCells.pkl','results/outputCells.sp')
hierPrimeCells.write_data('results/hierPrimeCells.pkl','spikes')
printPklSpikes('results/hierPrimeCells.pkl','results/hierPrimeCells.sp')
outputTimer.printResults('results/outputTimerCells')
printPklSpikes('results/outputTimerCells.pkl','results/outputTimerCells.sp')


sim.end()

