"""
assocStatic.py  This is the eighth step of the fifty item fifty association 
experiment.  

This uses compensatory_synapse learning rate at .01,
and Wmax_ is 2.0.  
The plastic cell parameters are also non-standard
'weight_std':20.0 undone,'tau_syn_ex':5.0,'g_L':20.0.

Read in the synapses from the hier.py.  They're static.  If you turn a CA on,
it fires.  If you prime the net, it makes its super category fire.

Have the operation and property cells from threeAssoc.py.

Learn four three way associations.  Test them by activating two works.
"""

import os #for printSynapticWeightsToFile
import pyNN.nest as sim
from pyNN.nest import native_synapse_type
from pyNN.nest import native_cell_type
import nealParams as nealParameters
from nealCoverClass import NealCoverFunctions

#constants
numAnimalCells = 200   #(animal (amphibian) 
                      #(fish shark salmon bass pike) #7
                      #(bird canary ostrich robin goose pigeon) 
                      #(mammal dog cat rat bear monkey human))  #20
numOperationCells = 50 #eats, breathes, is, has, can
numPropertyCells = 50  #(properties (can) move fly sing
                       #            (is) yellow (has) skin feathers gills)


import sys
import cPickle as pickle

def printPklSpikes(fileName,outFileName):
        fileHandle = open(fileName)
        outFileHandle = open(outFileName,'w')
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    outFileHandle.write(str( neuronNum) + " " + str(spikes[spike])+ "\n")
        outFileHandle.close()
        fileHandle.close()

def printSynapticWeights(synapses):
    weights = synapses.get(["weight"],format="array")
    print weights

def printSynapticWeightsToFile(synapses,fileName):
    fileHandle = open(fileName,'w')
    weights = synapses.get(["weight"],format="array")
    weightArray = weights[0]
    for preSynapticNeuronNumber in range (0,len(weightArray)):
        synapsesFromNeuron = weightArray[preSynapticNeuronNumber]
        fileHandle.write(str(preSynapticNeuronNumber))
        fileHandle.write(" ")
        fileHandle.write(str(synapsesFromNeuron))
        fileHandle.write("\n")
    fileHandle.write("\n")
    fileHandle.close()

def getTrainSpikeTimes(start, step, steps):
    spikesPerEpoch = 1
    inputSpikeTimes = []
    for numTrainSpikes in range (0,steps): 
        for spikesThisEpoch in range (0,spikesPerEpoch): 
            nextTime = start + (numTrainSpikes *step) + (spikesThisEpoch*5)
            inputSpikeTimes = inputSpikeTimes + [nextTime]

    return inputSpikeTimes

def createInput(startTime,numTrainingInputs,timeBetweenSteps):
    presentTimes = getTrainSpikeTimes(startTime+5,timeBetweenSteps,numTrainingInputs)
    presentArray = {'spike_times': [presentTimes]}
    primaryExciteSpikeGen=sim.Population(1,sim.SpikeSourceArray,presentArray)
    print presentTimes
    return primaryExciteSpikeGen

def createCAInputs():
    numInputs = 36
    timeBetweenEpochs = 250
    
    spikeSourceOpS1 = createInput(10,numInputs,timeBetweenEpochs) 
    spikeSourceOpS2 = createInput(60,numInputs,timeBetweenEpochs)
    spikeSourceOpS3 = createInput(110,numInputs,timeBetweenEpochs)
    spikeSourceOpS4 = createInput(160,numInputs,timeBetweenEpochs)
    spikeSourceOpS5 = createInput(210,numInputs,timeBetweenEpochs)
    stopSourceOp = createInput(50,numInputs*5,timeBetweenEpochs/5)

    propStart = numInputs*timeBetweenEpochs#9000 
    propSource1=createInput(propStart+10,numInputs,timeBetweenEpochs) 
    propSource2=createInput(propStart+60,numInputs,timeBetweenEpochs) 
    propSource3=createInput(propStart+110,numInputs,timeBetweenEpochs) 
    propSource4=createInput(propStart+160,numInputs,timeBetweenEpochs) 
    propSource5=createInput(propStart+210,numInputs,timeBetweenEpochs) 
    propStopSource = createInput(propStart+50,numInputs*5,timeBetweenEpochs/5)

    return [spikeSourceOpS1]+ [spikeSourceOpS2] + [spikeSourceOpS3]+ [
            spikeSourceOpS4] + [spikeSourceOpS5] + [stopSourceOp] + [
            propSource1] + [propSource2] + [propSource3] + [propSource4] + [
            propSource5] + [propStopSource] 

def makeFiveSynapsesFromSource(spikeSource,cells,toStart):
    inputWeight = 0.15
    connector = []
    for toOffset in range (0,5):
        toNeuron = toOffset +toStart
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

def makeTenSynapsesFromSource(spikeSource,cells,toStart):
    inputWeight = 0.05#3.0
    connector = []
    for toOffset in range (0,10):
        toNeuron = toOffset +toStart
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')


#During the association, keep the left hand side of the animal array inhibited.
def makeSynapsesForPreventingAnimalHier(cells,numCells,startTime,endTime):
    time = startTime
    timeArray = []
    while time < endTime:
        timeArray = timeArray + [time]
        time = time + 10

    presentArray = {'spike_times': timeArray}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    connector = []
    weight = -0.2
    numAnimals = numAnimalCells/10
    for animal in range (0,numAnimals):
        for toOffset in range (0,5):
            toNeuron  = (animal*10)+toOffset
            connector = connector + [(0,toNeuron,weight,1.0)]

    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

def sourceStopsPopulation(spikeSource,cells,popSize):
    inputWeight = -2.4 
    connector = []
    for toOffset in range (0,popSize):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

def sourceInhibitPopulation(spikeSource,cells,popSize,inputWeight):
    connector = []
    for toOffset in range (0,popSize):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

#Turn on an Animal and it's Operation
def makeTestAnimalOperation(animalCells,operationCells,propertyCells,timeStart,
                            animalStartNeuron,operationStartNeuron):
    #turn animal and operation on
    presentArray = {'spike_times': [timeStart]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.3
    connector = []
    for neuronNumber in range (0,5):
        toNeuron = animalStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, animalCells, connector,'excitatory')
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = operationStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, operationCells, connector,'excitatory')

    stopArray = {'spike_times': [timeStart+50]}
    stopSpikeSource= sim.Population(1,sim.SpikeSourceArray,stopArray)
    stopWeight = -0.4
    stopConnector = []
    for neuronNumber in range (0,numPropertyCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,propertyCells,stopConnector,       
                        'inhibitory')
    #won't work when they are different sizes
    neal.nealProjection(stopSpikeSource,operationCells,stopConnector,       
                        'inhibitory')
    stopConnector = []
    for neuronNumber in range (0,numAnimalCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,animalCells,stopConnector,       
                        'inhibitory')


#Turn on an Animal and property
def makeTestAnimalProperty(animalCells,operationCells,propertyCells,timeStart,
                            animalStartNeuron,propertyStartNeuron):
    #turn animal and property on
    presentArray = {'spike_times': [timeStart]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.3
    connector = []
    for neuronNumber in range (0,5):
        toNeuron = animalStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, animalCells, connector,'excitatory')
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = propertyStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, propertyCells, connector,'excitatory')

    stopArray = {'spike_times': [timeStart+50]}
    stopSpikeSource= sim.Population(1,sim.SpikeSourceArray,stopArray)
    stopWeight = -0.4
    stopConnector = []
    for neuronNumber in range (0,numPropertyCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,propertyCells,stopConnector,       
                        'inhibitory')
    #won't work when they are different sizes
    neal.nealProjection(stopSpikeSource,operationCells,stopConnector,       
                        'inhibitory')
    stopConnector = []
    for neuronNumber in range (0,numAnimalCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,animalCells,stopConnector,       
                        'inhibitory')


def makeTestOperationProperty(animalCells,operationCells,propertyCells,
                        timeStart,operationStartNeuron,propertyStartNeuron):
    #turn animal and operation on
    presentArray = {'spike_times': [timeStart]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.3
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = operationStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, operationCells, connector,'excitatory')
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = propertyStartNeuron + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, propertyCells, connector,'excitatory')

    stopArray = {'spike_times': [timeStart+50]}
    stopSpikeSource= sim.Population(1,sim.SpikeSourceArray,stopArray)
    stopWeight = -0.4
    stopConnector = []
    for neuronNumber in range (0,numPropertyCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,propertyCells,stopConnector,       
                        'inhibitory')
    #won't work when they are different sizes
    neal.nealProjection(stopSpikeSource,operationCells,stopConnector,       
                        'inhibitory')
    stopConnector = []
    for neuronNumber in range (0,numAnimalCells):
        toNeuron = neuronNumber
        stopConnector = stopConnector + [
                (0,toNeuron,stopWeight,nealParameters.DELAY)]
    neal.nealProjection(stopSpikeSource,animalCells,stopConnector,       
                        'inhibitory')


def makeTestSynapses(animalCells,operationCells,propertyCells):
    makeTestAnimalOperation(animalCells,operationCells,propertyCells,26000,
                            85,20) #canary is -> yellow
    makeTestOperationProperty(animalCells,operationCells,propertyCells,26100,
                              40,10) #can fly -> bird
    makeTestAnimalProperty(animalCells,operationCells,propertyCells,26200,
                            5,40) #animal skin -> has
    makeTestAnimalOperation(animalCells,operationCells,propertyCells,26300,
                            25,0) #fish gills -> breathesWith


def makePrimeSourceSynapses(cells,numCells):
    #presentArray = {'spike_times': [145000]}
    presentArray = {'spike_times': [501]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.02 #.05 .01
    connector = []
    for toNeuron in range (0,numCells):
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

def readSynapses(cells,numCells):
    connector = []
    preSynapticNeuronNumber = -1
    postSynapticNeuronNumber = 0
    needNeuronNumber = True
    inputFile = open("results/storedDoubleAnimalWeights.txt",'r')
    inputLine = inputFile.readline()
    #while inputLine != "":
    while (len(inputLine) > 1): #go until a blank line
        inputLine = inputLine.strip() #remove the \n
        #print inputLine
        inputs = inputLine.split(' ')
        numInputs = len(inputs)
        currentInput = 0
        if (needNeuronNumber):
            needNeuronNumber = False
            postSynapticNeuronNumber = 0
            temp = int(inputs[0])
            preSynapticNeuronNumber = preSynapticNeuronNumber + 1
            if (preSynapticNeuronNumber != temp):
                    print 'bad neuron number'
            currentInput = 3

        #see if the line has the last synapse
        if (']' in inputLine):
            needNeuronNumber = True
            #remove the ] from the last input
            lastInput =  inputs[numInputs-1]
            lastInput = lastInput[0:len(lastInput)-1]
            inputs[numInputs-1] = lastInput
        
        #process one line
        while (currentInput < numInputs):
            synapticWeight = float(inputs[currentInput])
            #print preSynapticNeuronNumber,postSynapticNeuronNumber,synapticWeight
            if (synapticWeight > 0.00001):
            #if ((synapticWeight > 0.00001) and (preSynapticNeuronNumber != postSynapticNeuronNumber)):
                    connector = connector + [(preSynapticNeuronNumber,postSynapticNeuronNumber,synapticWeight,1.0)]
            postSynapticNeuronNumber = postSynapticNeuronNumber + 1
            currentInput = currentInput + 3

        inputLine = inputFile.readline()
    inputFile.close()
    
    print "synapses " + str(len(connector))
    connectorList = sim.FromListConnector(connector)
    intraAnimalSynapses = sim.Projection(AnimalCells,AnimalCells, 
                connectorList, sim.StaticSynapse()) 

    return intraAnimalSynapses

def makeSynapsesFromSpikeSources(spikeSources,cells,popSize,otherCells,otherPopSize):
    s1OperationSource = spikeSources[0] 
    s2OperationSource = spikeSources[1]
    s3OperationSource = spikeSources[2]
    s4OperationSource = spikeSources[3]
    s5OperationSource = spikeSources[4]
    stopOperationSource = spikeSources[5]

    makeTenSynapsesFromSource(s1OperationSource,cells,0)
    makeTenSynapsesFromSource(s2OperationSource,cells,10)
    makeTenSynapsesFromSource(s3OperationSource,cells,20)
    makeTenSynapsesFromSource(s4OperationSource,cells,30)
    makeTenSynapsesFromSource(s5OperationSource,cells,40)
    sourceStopsPopulation(stopOperationSource,cells,popSize)
    #sourceInhibitPopulation(stopOperationSource,otherCells,otherPopSize,-0.05)

def makeAssocSynapsesFromSpikeSources(animalCells,animalPopSize,
                                      operationCells,operationPopSize,
                                      propertyCells,propertyPopSize):
        
    presentArray1 = {'spike_times': [21000,21100,21200,21300,21400,21500]}
    assocSource1=sim.Population(1,sim.SpikeSourceArray,presentArray1)
    presentArray2 = {'spike_times': [22000,22100,22200,22300,22400,22500]}
    assocSource2=sim.Population(1,sim.SpikeSourceArray,presentArray2)
    presentArray3 = {'spike_times': [23000,23100,23200,23300,23400,23500]}
    assocSource3=sim.Population(1,sim.SpikeSourceArray,presentArray3)
    presentArray4 = {'spike_times': [24000,24100,24200,24300,24400,24500]}
    assocSource4=sim.Population(1,sim.SpikeSourceArray,presentArray4)
    presentArray5 = {'spike_times': [25000,35100,35200,35300,35400,35500]}

    stopArray = {'spike_times': [21050,21150,21250,21350,21450,21550, 
                                 22050,22150,22250,22350,22450,22550,
                                 23050,23150,23250,23350,23450,23550,
                                 24050,24150,24250,24350,24450,24550]}
                                 #25050,35150,35250,35350,35450,35550]}
    stopCellsSource=sim.Population(1,sim.SpikeSourceArray,stopArray)

    makeSynapsesForPreventingAnimalHier(animalCells,numAnimalCells,21000,25000)

    #canary is yellow
    makeFiveSynapsesFromSource(assocSource1,animalCells,85)
    makeTenSynapsesFromSource(assocSource1,operationCells,20)
    makeTenSynapsesFromSource(assocSource1,propertyCells,30)

    #bird can fly
    makeFiveSynapsesFromSource(assocSource2,animalCells,75)
    makeTenSynapsesFromSource(assocSource2,operationCells,40)
    makeTenSynapsesFromSource(assocSource2,propertyCells,10)

    #animal has skin
    makeFiveSynapsesFromSource(assocSource3,animalCells,5)
    makeTenSynapsesFromSource(assocSource3,operationCells,30)
    makeTenSynapsesFromSource(assocSource3,propertyCells,40) 

    #fish breathes withGills
    makeFiveSynapsesFromSource(assocSource4,animalCells,25)
    makeTenSynapsesFromSource(assocSource4,operationCells,10)
    makeTenSynapsesFromSource(assocSource4,propertyCells,0) #undone not above

    #bass eats worms
    #makeFiveSynapsesFromSource(assocSource4,animalCells,55)
    #makeTenSynapsesFromSource(assocSource4,operationCells,0)
    #makeTenSynapsesFromSource(assocSource4,propertyCells,20) #undone not above

    sourceInhibitPopulation(stopCellsSource,animalCells,animalPopSize,-2.5)
    sourceInhibitPopulation(stopCellsSource,operationCells,operationPopSize,
                            -2.5)
    sourceInhibitPopulation(stopCellsSource,propertyCells,propertyPopSize,-2.5) 

def getWellConnectedConnectorList(numCells):
    connector = []
    weight = 0.0001
    
    #make all connections except self connections
    for fromOffset in range (0,numCells):
        for toOffset in range (0,numCells):
            if (fromOffset != toOffset):
                connector = connector + [(fromOffset,toOffset,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList

def getCrossConnectedConnectorList(numFromCells,numToCells):
    connector = []
    weight = 0.1
    
    #make all connections except self connections
    for fromOffset in range (0,numFromCells):
        for toOffset in range (0,numToCells):
            #don't make well connected but 40%
            if (((fromOffset+toOffset) %10 ) < 2):
                if (fromOffset != toOffset):
                    connector = connector + [(fromOffset,toOffset,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList

#Get a connector list for cell numbers from the right hand side
#of the hierarchy.
def getFromAnimalConnectorList(numAnimalCells,numToCells):
    connector = []
    weight = 0.1
    numAnimals = numAnimalCells/10
    for animal in range (0,numAnimals):
        for fromOffset in range (5,10):
            for toOffset in range (0,numToCells):
                #don't make well connected but 40% undone
                fromNeuron  = (animal*10)+fromOffset
                connector = connector + [(fromNeuron,toOffset,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList

#get a connector list for cell numbers for the right hand side
#of the hierarchy
def getToAnimalConnectorList(numFromCells,numAnimalCells):
    connector = []
    weight = 0.1
    numAnimals = numAnimalCells/10
    for fromOffset in range (0,numFromCells):
        for animal in range (0,numAnimals):
            for toOffset in range (5,10):
                #don't make well connected but 40% undone
                toNeuron  = (animal*10)+toOffset
                connector = connector + [(fromOffset,toNeuron,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList


def makePlasticConnections():
    compense_synapse_type = native_synapse_type('compensatory_synapse')
    CompenseSynapseModel = compense_synapse_type() 

    #make connections within one population (intra)
    intraOperationConnList = getWellConnectedConnectorList(numOperationCells)
    intraOperationSynapses = sim.Projection(OperationCells,OperationCells, 
                intraOperationConnList, synapse_type = CompenseSynapseModel) 

    intraPropertyConnectorList = getWellConnectedConnectorList(numPropertyCells)
    intraPropertySynapses = sim.Projection(PropertyCells,PropertyCells, 
                intraPropertyConnectorList, synapse_type = CompenseSynapseModel) 
    #make connections between populations
    animalToOperationConnectorList = getFromAnimalConnectorList(
        numAnimalCells,numOperationCells)
    animalToOperationSynapses = sim.Projection(AnimalCells,OperationCells, 
            animalToOperationConnectorList, synapse_type=CompenseSynapseModel) 

    animalToPropertyConnectorList = getFromAnimalConnectorList(
        numAnimalCells,numPropertyCells)
    animalToPropertySynapses = sim.Projection(AnimalCells,PropertyCells, 
            animalToPropertyConnectorList, synapse_type=CompenseSynapseModel) 

    operationToAnimalConnectorList = getToAnimalConnectorList(
        numOperationCells,numAnimalCells)
    operationToAnimalSynapses = sim.Projection(OperationCells,AnimalCells,
            operationToAnimalConnectorList, synapse_type=CompenseSynapseModel) 

    operationToPropertyConnectorList = getCrossConnectedConnectorList(
        numOperationCells,numPropertyCells)
    operationToPropertySynapses = sim.Projection(OperationCells,PropertyCells,
        operationToPropertyConnectorList, synapse_type=CompenseSynapseModel) 

    propertyToAnimalConnectorList = getToAnimalConnectorList(
        numPropertyCells,numAnimalCells)
    propertyToAnimalSynapses = sim.Projection(PropertyCells,AnimalCells,
            propertyToAnimalConnectorList, synapse_type=CompenseSynapseModel) 

    propertyToObjectConnectorList = getCrossConnectedConnectorList(
        numPropertyCells,numOperationCells)
    propertyToObjectSynapses = sim.Projection(PropertyCells,OperationCells,
            propertyToObjectConnectorList, synapse_type=CompenseSynapseModel) 

    return [intraOperationSynapses]+[
        intraPropertySynapses]+ [animalToOperationSynapses] + [
        animalToPropertySynapses] + [operationToAnimalSynapses] + [
        operationToPropertySynapses]+ [propertyToAnimalSynapses] + [
        propertyToObjectSynapses] 

def printPlasticSynapticWeights(synapseSets):
    intraOperationSynapses = synapseSets[0]
    intraPropertySynapses = synapseSets[1]
    animalToOperationSynapses = synapseSets[2]
    animalToPropertySynapses = synapseSets[3]
    operationToAnimalSynapses = synapseSets[4]
    operationToPropertySynapses = synapseSets[5]
    propertyToAnimalSynapses = synapseSets[6]
    propertyToOperationSynapses = synapseSets[7]

    printSynapticWeightsToFile(intraPropertySynapses,
                               "results/propertyWeights.txt")
    printSynapticWeightsToFile(intraOperationSynapses,
                               "results/operationWeights.txt")
    printSynapticWeightsToFile(animalToOperationSynapses,
                               "results/animalToOperationWeights.txt")
    printSynapticWeightsToFile(animalToPropertySynapses,
                               "results/animalToPropertyWeights.txt")
    printSynapticWeightsToFile(operationToAnimalSynapses,
                               "results/operationToAnimalWeights.txt")
    printSynapticWeightsToFile(operationToPropertySynapses,
                               "results/operationToPropertyWeights.txt")
    printSynapticWeightsToFile(propertyToAnimalSynapses,
                               "results/propertyToAnimalWeights.txt")
    printSynapticWeightsToFile(propertyToOperationSynapses,
                               "results/propertyToOperationWeights.txt")

#--main-
sim.nest.Install("chrismodule")
CompenseIFCondExpNeuronType = native_cell_type('compenseIF_cond_exp')
print sim.nest.GetDefaults('compenseIF_cond_exp')
neal = NealCoverFunctions(nealParameters.simulator)

sim.setup(timestep=1.0, max_delay=1.0)

spikeSources = createCAInputs()

#Create the Neurons
AnimalCells = sim.Population(numAnimalCells,CompenseIFCondExpNeuronType, 
        {'weight_std':21.0,'tau_syn_ex':5.0,'g_L':20.0,'tau_syn_in' : 2.0})
AnimalCells.record(['spikes'])
OperationCells = sim.Population(numOperationCells,CompenseIFCondExpNeuronType, 
                                {'weight_std':20.0,'tau_syn_ex':5.0,'g_L':20.0})
OperationCells.record(['spikes'])
PropertyCells = sim.Population(numPropertyCells,CompenseIFCondExpNeuronType, 
                                {'weight_std':20.0,'tau_syn_ex':5.0,'g_L':20.0})
PropertyCells.record(['spikes'])

#Make the synapses
#from spike sources
makeSynapsesFromSpikeSources(spikeSources,OperationCells,numOperationCells,AnimalCells,numAnimalCells)
makeSynapsesFromSpikeSources(spikeSources[6:12],PropertyCells,numPropertyCells,OperationCells,numOperationCells)
durationForCAs = 20000
makeAssocSynapsesFromSpikeSources(AnimalCells,numAnimalCells,OperationCells,numOperationCells,PropertyCells,numPropertyCells)
durationForAssocation = 6000

makeTestSynapses(AnimalCells,OperationCells,PropertyCells)
#makePrimeSourceSynapses(AnimalCells,numAnimalCells)

intraAnimalSynapses = readSynapses(AnimalCells,numAnimalCells)
synapseSets = makePlasticConnections()

#18000 for training operation and property CAs
durationForTest = 1000
duration = durationForCAs + durationForAssocation + durationForTest  
print duration
sim.run(duration)

print AnimalCells[75].weight_std
print AnimalCells[75].weighted_spikes_ex
print OperationCells[40].weight_std
print OperationCells[40].weighted_spikes_ex


AnimalCells.write_data('results/animalCells.pkl','spikes')
OperationCells.write_data('results/operationCells.pkl','spikes')
PropertyCells.write_data('results/propertyCells.pkl','spikes')
printPklSpikes('results/animalCells.pkl','results/animalCells.sp')
printPklSpikes('results/operationCells.pkl','results/operationCells.sp')
printPklSpikes('results/propertyCells.pkl','results/propertyCells.sp')

printPlasticSynapticWeights(synapseSets)
printSynapticWeightsToFile(intraAnimalSynapses,"results/animalWeights.txt")

sim.end()

