"""
doubleHier.py  this is the sixth step of the fifty item fifty association 
experiment.  It's a second version of hier.py.   Each node in the
hierarchy has 10 cells; 5 for inheritence; 5 for associations.

Work with the compensatory_synapse learning rate at .01,
and Wmax_ is 2.0.  The plastic cell parameters are also non-standard
'weight_std':15.0,'tau_syn_ex':5.0,'g_L':20.0.

The goal is to get the individual items persisting. It should
also spread up the hierarchy when the associations are primed.  
"""

import os #for printSynapticWeightsToFile
import pyNN.nest as sim
from pyNN.nest import native_synapse_type
from pyNN.nest import native_cell_type
import nealParams as nealParameters
from nealCoverClass import NealCoverFunctions

#constants
numAnimalCells = 200   #(animal (amphibian) 
                      #(fish shark salmon bass pike) #7
                      #(bird canary ostrich robin goose pigeon) 
                      #(mammal dog cat rat bear monkey human))  #20

import sys
import cPickle as pickle

numInputs = 30
epochLength = 5000

def printPklSpikes(fileName,outFileName):
        fileHandle = open(fileName)
        outFileHandle = open(outFileName,'w')
        neoObj = pickle.load(fileHandle)
        segments = neoObj.segments
        segment = segments[0]
        spikeTrains = segment.spiketrains
        neurons = len(spikeTrains)
        for neuronNum in range (0,neurons):
            if (len(spikeTrains[neuronNum])>0):
                spikes = spikeTrains[neuronNum]
                for spike in range (0,len(spikes)):
                    outFileHandle.write(str( neuronNum) + " " + str(spikes[spike])+ "\n")
        outFileHandle.close()
        fileHandle.close()

def printSynapticWeights(synapses):
    weights = synapses.get(["weight"],format="array")
    print weights

def printSynapticWeightsToFile(synapses,fileName):
    fileHandle = open(fileName,'w')
    weights = synapses.get(["weight"],format="array")
    weightArray = weights[0]
    for preSynapticNeuronNumber in range (0,len(weightArray)):
        synapsesFromNeuron = weightArray[preSynapticNeuronNumber]
        fileHandle.write(str(preSynapticNeuronNumber))
        fileHandle.write(" ")
        fileHandle.write(str(synapsesFromNeuron))
        fileHandle.write("\n")
    fileHandle.write("\n")
    fileHandle.close()

#Get a series of spikes starting at start, and going by step.  There
#are steps of them.
def getTrainSpikeTimes(start, step, steps):
    extraTime = 0
    inputSpikeTimes = []
    for numTrainSpikes in range (0,steps): 
        nextTime = start + (numTrainSpikes *step) + extraTime
        inputSpikeTimes = inputSpikeTimes + [nextTime]

    #print inputSpikeTimes
    return inputSpikeTimes

def createInput(startTime,numTrainingInputs,timeBetweenSteps):
    presentTimes = getTrainSpikeTimes(startTime+5,timeBetweenSteps,numTrainingInputs)
    presentArray = {'spike_times': [presentTimes]}
    primaryExciteSpikeGen=sim.Population(1,sim.SpikeSourceArray,presentArray)
    #print presentTimes
    return primaryExciteSpikeGen

def createCAInputs():
    timeBetweenEpochs = epochLength
    CASource1 = createInput(10,numInputs,timeBetweenEpochs) 
    CASource2 = createInput(60,numInputs,timeBetweenEpochs) 
    CASource3 = createInput(110,numInputs,timeBetweenEpochs) 
    CASource4 = createInput(160,numInputs,timeBetweenEpochs) 
    CASource5 = createInput(210,numInputs,timeBetweenEpochs) 
    CASource6 = createInput(260,numInputs,timeBetweenEpochs) 
    CASource7 = createInput(310,numInputs,timeBetweenEpochs) 
    CASource8 = createInput(360,numInputs,timeBetweenEpochs) 
    CASource9 = createInput(410,numInputs,timeBetweenEpochs) 
    CASource10 = createInput(460,numInputs,timeBetweenEpochs) 
    CASource11 = createInput(510,numInputs,timeBetweenEpochs) 
    CASource12 = createInput(560,numInputs,timeBetweenEpochs) 
    CASource13 = createInput(610,numInputs,timeBetweenEpochs) 
    CASource14 = createInput(660,numInputs,timeBetweenEpochs) 
    CASource15 = createInput(760,numInputs,timeBetweenEpochs) 
    CASource16 = createInput(810,numInputs,timeBetweenEpochs) 
    CASource17 = createInput(860,numInputs,timeBetweenEpochs) 
    CASource18 = createInput(910,numInputs,timeBetweenEpochs) 
    CASource19 = createInput(960,numInputs,timeBetweenEpochs) 
    CASource20 = createInput(1010,numInputs,timeBetweenEpochs) 
    AssocSource1 = createInput(1060,numInputs,timeBetweenEpochs) 
    AssocSource2 = createInput(1110,numInputs,timeBetweenEpochs) 
    AssocSource3 = createInput(1160,numInputs,timeBetweenEpochs) 
    AssocSource4 = createInput(1210,numInputs,timeBetweenEpochs) 
    AssocSource5 = createInput(1260,numInputs,timeBetweenEpochs) 
    AssocSource6 = createInput(1310,numInputs,timeBetweenEpochs) 
    AssocSource7 = createInput(1360,numInputs,timeBetweenEpochs) 
    AssocSource8 = createInput(1410,numInputs,timeBetweenEpochs) 
    AssocSource9 = createInput(1460,numInputs,timeBetweenEpochs) 
    AssocSource10 = createInput(1510,numInputs,timeBetweenEpochs) 
    AssocSource11 = createInput(1560,numInputs,timeBetweenEpochs) 
    AssocSource12 = createInput(1610,numInputs,timeBetweenEpochs) 
    AssocSource13 = createInput(1660,numInputs,timeBetweenEpochs) 
    AssocSource14 = createInput(1710,numInputs,timeBetweenEpochs) 
    AssocSource15 = createInput(1760,numInputs,timeBetweenEpochs) 
    AssocSource16 = createInput(1810,numInputs,timeBetweenEpochs) 
    AssocSource17 = createInput(1860,numInputs,timeBetweenEpochs) 
    AssocSource18 = createInput(1910,numInputs,timeBetweenEpochs) 
    AssocSource19 = createInput(1960,numInputs,timeBetweenEpochs) 

    return [CASource1]+ [CASource2] + [CASource3] + [
            CASource4]+ [CASource5] + [CASource6] + [
            CASource7]+ [CASource8] + [CASource9] + [
            CASource10]+ [CASource11] + [CASource12] + [
            CASource13]+ [CASource14] + [CASource15] + [
            CASource16]+ [CASource17] + [CASource18] + [
            CASource19]+ [CASource20] + [
            AssocSource1]+ [AssocSource2] + [AssocSource3] + [
            AssocSource4]+ [AssocSource5] + [AssocSource6] + [
            AssocSource7]+ [AssocSource8] + [AssocSource9] + [
            AssocSource10]+ [AssocSource11] + [AssocSource12] + [
            AssocSource13]+ [AssocSource14] + [AssocSource15] + [
            AssocSource16]+ [AssocSource7] + [AssocSource18] + [
            AssocSource19]

def makeFiveSynapsesFromSource(spikeSource,cells,toStart):
    inputWeight = 0.05#3.0
    connector = []
    for toOffset in range (0,5):
        toNeuron = toOffset +toStart
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

def makeTenSynapsesFromSource(spikeSource,cells,toStart):
    inputWeight = 0.05#3.0
    connector = []
    for toOffset in range (0,10):
        toNeuron = toOffset +toStart
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

def makeSynapsesFromSpikeSourcesToCAs(cells,popSize):
    spikeSources = createCAInputs()
    for CA in range (0,20):
            makeTenSynapsesFromSource(spikeSources[CA],cells,CA*10)

def makeSynapsesFromSpikeSourcesToHiers(cells,popSize):
    spikeSources = createCAInputs()

    #animal amphibian
    makeFiveSynapsesFromSource(spikeSources[20],cells,0)
    makeFiveSynapsesFromSource(spikeSources[20],cells,10)

    #animal fish
    makeFiveSynapsesFromSource(spikeSources[21],cells,0)
    makeFiveSynapsesFromSource(spikeSources[21],cells,20)

    #fish shark
    makeFiveSynapsesFromSource(spikeSources[22],cells,20)
    makeFiveSynapsesFromSource(spikeSources[22],cells,30)

    #fish salmon
    makeFiveSynapsesFromSource(spikeSources[23],cells,20)
    makeFiveSynapsesFromSource(spikeSources[23],cells,40)

    #fish bass
    makeFiveSynapsesFromSource(spikeSources[24],cells,20)
    makeFiveSynapsesFromSource(spikeSources[24],cells,50)

    #fish pike
    makeFiveSynapsesFromSource(spikeSources[25],cells,20)
    makeFiveSynapsesFromSource(spikeSources[25],cells,60)

    #animal bird
    makeFiveSynapsesFromSource(spikeSources[26],cells,0)
    makeFiveSynapsesFromSource(spikeSources[26],cells,70)

    #bird canary
    makeFiveSynapsesFromSource(spikeSources[27],cells,70)
    makeFiveSynapsesFromSource(spikeSources[27],cells,80)

    #bird ostrich
    makeFiveSynapsesFromSource(spikeSources[27],cells,70)
    makeFiveSynapsesFromSource(spikeSources[27],cells,90)

    #bird robin
    makeFiveSynapsesFromSource(spikeSources[28],cells,70)
    makeFiveSynapsesFromSource(spikeSources[28],cells,100)

    #bird goose
    makeFiveSynapsesFromSource(spikeSources[29],cells,70)
    makeFiveSynapsesFromSource(spikeSources[29],cells,110)

    #bird pigeon 30
    makeFiveSynapsesFromSource(spikeSources[30],cells,70)
    makeFiveSynapsesFromSource(spikeSources[30],cells,120)

    #animal mammal
    makeFiveSynapsesFromSource(spikeSources[31],cells,0)
    makeFiveSynapsesFromSource(spikeSources[31],cells,130)

    #mammal dog
    makeFiveSynapsesFromSource(spikeSources[32],cells,130)
    makeFiveSynapsesFromSource(spikeSources[32],cells,140)

    #mammal cat
    makeFiveSynapsesFromSource(spikeSources[33],cells,130)
    makeFiveSynapsesFromSource(spikeSources[33],cells,150)

    #mammal rat
    makeFiveSynapsesFromSource(spikeSources[34],cells,130)
    makeFiveSynapsesFromSource(spikeSources[34],cells,160)

    #mammal bear  35
    makeFiveSynapsesFromSource(spikeSources[35],cells,130)
    makeFiveSynapsesFromSource(spikeSources[35],cells,170)

    #mammal monkey
    makeFiveSynapsesFromSource(spikeSources[36],cells,130)
    makeFiveSynapsesFromSource(spikeSources[36],cells,180)

    #mammal human
    makeFiveSynapsesFromSource(spikeSources[37],cells,130)
    makeFiveSynapsesFromSource(spikeSources[37],cells,190)


def getStopTrainSpikeTimes(start,steps,stepLength,skipStart):
    inputSpikeTimes = []
    for epoch in range (0,numInputs):
        for numTrainSpikes in range (0,steps): 
            skipTime = epoch * epochLength
            nextTime = start + (numTrainSpikes *stepLength) + skipTime
            inputSpikeTimes = inputSpikeTimes + [nextTime]

    #print inputSpikeTimes
    return inputSpikeTimes

def makeStopSynapsesFromSpikeSources(cells, popSize):
    #get the CA formation times 
    cAStops = 20 + 19 + 2
    stopTimes = getStopTrainSpikeTimes(0,cAStops,50,2000)
    presentArray = {'spike_times': [stopTimes]}
    inhibitSpikeGen=sim.Population(1,sim.SpikeSourceArray,presentArray)
    sourceStopsPopulation(inhibitSpikeGen,cells,popSize)



def sourceStopsPopulation(spikeSource,cells,popSize):
    inputWeight = -0.4 
    connector = []
    for toOffset in range (0,popSize):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

def sourceInhibitPopulation(spikeSource,cells,popSize,inputWeight):
    connector = []
    for toOffset in range (0,popSize):
        toNeuron = toOffset
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'inhibitory')

#create a spike source for every neuron.  Have each spike once every
#refreshEpoch
#make one synapse from each of these spike sources to each cell.
#the idea is that it will fire just the one cell.
def makeOneSpikeSourceSynapses(cells,popSize):
    startTime = 2050
    for neuronNumber in range (0,popSize):
        inputSpikeTimes = []
        for epoch in range (0,numInputs):
            nextTime = startTime + (neuronNumber*11) + (epoch*epochLength)
            inputSpikeTimes = inputSpikeTimes + [nextTime]

        presentArray = {'spike_times': [inputSpikeTimes]}
        spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
        inputWeight = 0.03
        connector = []
        connector = connector + [(0,neuronNumber,inputWeight,
                                  nealParameters.DELAY)]
        neal.nealProjection(spikeSource, cells, connector,'excitatory')

def makeTestSourceSynapses(cells):
    time = (numInputs*epochLength) + 500
    presentArray = {'spike_times': [time]}
    spikeSource= sim.Population(1,sim.SpikeSourceArray,presentArray)
    inputWeight = 0.05
    connector = []
    for neuronNumber in range (0,5):
        toNeuron = 60 + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource, cells, connector,'excitatory')

    time = (numInputs*epochLength) + 1000
    presentArray = {'spike_times': [time]}
    spikeSource2= sim.Population(1,sim.SpikeSourceArray,presentArray)
    connector = []
    for neuronNumber in range (0,10):
        toNeuron = 60 + neuronNumber
        connector = connector + [(0,toNeuron,inputWeight,nealParameters.DELAY)]
    neal.nealProjection(spikeSource2, cells, connector,'excitatory')

def getWellConnectedConnectorList(numCells):
    connector = []
    weight = 0.0001
    
    #make all connections except self connections
    for fromOffset in range (0,numCells):
        for toOffset in range (0,numCells):
            if (fromOffset != toOffset):
                connector = connector + [(fromOffset,toOffset,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList

def inCA(fromNeuron,toNeuron):
    return False
    if (fromNeuron == toNeuron): 
        return False
    fromCA = fromNeuron/5
    toCA = toNeuron/5
    if (fromCA == toCA):
        return True
    return False

def getWellConnectedCAConnectorList(numCells):
    connector = []
    weight = 0.0001
    CAWeight = 0.005
    
    #make all connections except self connections
    for fromOffset in range (0,numCells):
        for toOffset in range (0,numCells):
            if (inCA(fromOffset,toOffset)):
                connector = connector + [(fromOffset,toOffset,CAWeight,1.0)]
            #elif (fromOffset != toOffset):
            connector = connector + [(fromOffset,toOffset,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList
    
def getStaticWellConnectedConnectorList(numCells,weight):
    connector = []
    
    #make all connections except self connections
    for CA in range (0,numCells/10):
        for fromOffset in range (0,10):
            fromNeuron = (CA*10)+fromOffset
            for toOffset in range (0,10):
                if (fromOffset != toOffset):
                    toNeuron = (CA*10)+toOffset
                    connector = connector + [(fromNeuron,toNeuron,weight,1.0)]

    connectorList = sim.FromListConnector(connector)
    return connectorList

def makeStaticConnections():
    wellConnectedWeight = 0.0015#2.2 
    crossConnectionWeight = 0.002#0.03 #.03 - 0.05
    intraAnimalConnectorList = getStaticWellConnectedConnectorList(numAnimalCells,wellConnectedWeight)
    intraAnimalSynapses = sim.Projection(AnimalCells,AnimalCells, 
                intraAnimalConnectorList, sim.StaticSynapse()) 

    intraOperationConnList = getStaticWellConnectedConnectorList(numOperationCells,wellConnectedWeight)
    intraOperationSynapses = sim.Projection(OperationCells,OperationCells, 
                intraOperationConnList, sim.StaticSynapse())  

    intraPropertyConnectorList = getStaticWellConnectedConnectorList(numPropertyCells,getWellConnectedWeight)
    intraPropertySynapses = sim.Projection(PropertyCells,PropertyCells, 
                intraPropertyConnectorList, sim.StaticSynapse()) 
 
    #make connections between populations
    animalToOperationConnectorList = getStaticCrossConnectedConnectorList(
        numAnimalCells,numOperationCells,crossConnectionWeight)
    animalToOperationSynapses = sim.Projection(AnimalCells,OperationCells, 
            animalToOperationConnectorList, sim.StaticSynapse())  

    animalToPropertyConnectorList = getStaticCrossConnectedConnectorList(
        numAnimalCells,numPropertyCells,crossConnectionWeight)
    animalToPropertySynapses = sim.Projection(AnimalCells,PropertyCells, 
            animalToPropertyConnectorList, sim.StaticSynapse()) 

    operationToAnimalConnectorList = getStaticCrossConnectedConnectorList(
        numOperationCells,numAnimalCells,crossConnectionWeight)
    operationToAnimalSynapses = sim.Projection(OperationCells,AnimalCells,
            operationToAnimalConnectorList, sim.StaticSynapse()) 

    operationToPropertyConnectorList = getStaticCrossConnectedConnectorList(
        numOperationCells,numPropertyCells,crossConnectionWeight)
    operationToPropertySynapses = sim.Projection(OperationCells,PropertyCells,
        operationToPropertyConnectorList, sim.StaticSynapse()) 

    propertyToAnimalConnectorList = getStaticCrossConnectedConnectorList(
        numPropertyCells,numAnimalCells,crossConnectionWeight)
    propertyToAnimalSynapses = sim.Projection(PropertyCells,AnimalCells,
            propertyToAnimalConnectorList,sim.StaticSynapse()) 

    propertyToObjectConnectorList = getStaticCrossConnectedConnectorList(
        numPropertyCells,numOperationCells,crossConnectionWeight)
    propertyToObjectSynapses = sim.Projection(PropertyCells,OperationCells,
            propertyToObjectConnectorList,sim.StaticSynapse()) 

    return [intraOperationSynapses]+[intraAnimalSynapses]+ [
        intraPropertySynapses]+ [animalToOperationSynapses] + [
        animalToPropertySynapses] + [operationToAnimalSynapses] + [
        operationToPropertySynapses]+ [propertyToAnimalSynapses] + [
        propertyToObjectSynapses] 

def makePlasticConnections():
    compense_synapse_type = native_synapse_type('compensatory_synapse')
    CompenseSynapseModel = compense_synapse_type() 

    #make connections within one population (intra)
    #intraAnimalConnectorList = getWellConnectedConnectorList(numAnimalCells)
    intraAnimalConnectorList = getWellConnectedCAConnectorList(numAnimalCells)
    intraAnimalSynapses = sim.Projection(AnimalCells,AnimalCells, 
                intraAnimalConnectorList, synapse_type = CompenseSynapseModel) 

    return [intraAnimalSynapses]

#--main-
sim.nest.Install("chrismodule")
CompenseIFCondExpNeuronType = native_cell_type('compenseIF_cond_exp')
print sim.nest.GetDefaults('compenseIF_cond_exp')
neal = NealCoverFunctions(nealParameters.simulator)

sim.setup(timestep=1.0, max_delay=1.0)

#spikeSources = createInputs()

#Create the Neurons
AnimalCells = sim.Population(numAnimalCells,CompenseIFCondExpNeuronType, 
        {'weight_std':15.0,'tau_syn_ex':5.0,'g_L':20.0,'tau_syn_in' : 2.0})
AnimalCells.record(['spikes'])

#make the synapses from spike sources
makeSynapsesFromSpikeSourcesToCAs(AnimalCells,numAnimalCells)
makeSynapsesFromSpikeSourcesToHiers(AnimalCells,numAnimalCells)
makeStopSynapsesFromSpikeSources(AnimalCells,numAnimalCells)
makeOneSpikeSourceSynapses(AnimalCells,numAnimalCells)
makeTestSourceSynapses(AnimalCells)

synapseSets = makePlasticConnections()
#synapseSets = makeStaticConnections()

intraAnimalSynapses = synapseSets[0]

duration = numInputs*epochLength #train
#duration = duration + 1200 #test
print duration 
sim.run(duration)

printSynapticWeightsToFile(intraAnimalSynapses,"results/animalWeights.txt")

print AnimalCells[0].weight_std
print AnimalCells[0].weighted_spikes_ex
print AnimalCells[5].weighted_spikes_ex
print AnimalCells[9].weighted_spikes_ex
print AnimalCells[10].weighted_spikes_ex
print AnimalCells[15].weighted_spikes_ex
print AnimalCells[20].weighted_spikes_ex
print AnimalCells[30].weighted_spikes_ex
print AnimalCells[40].weighted_spikes_ex
#print AnimalCells[100].weighted_spikes_ex
#print AnimalCells[110].weighted_spikes_ex

AnimalCells.write_data('results/animalCells.pkl','spikes')
printPklSpikes('results/animalCells.pkl','results/animalCells.sp')

sim.end()

